#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_SYMBOLS 2000
#define PIC_ADDR    0x140 // Default PIC table address

int addr = 0;
int debugflag = 0;

char type[MAX_SYMBOLS];
int value[MAX_SYMBOLS];
char *name[MAX_SYMBOLS];
int numsym = 0;
char bigbuf[MAX_SYMBOLS*32];
char *bufptr = bigbuf;
int objnum = 0;
int varcount = 0;
int pictableaddr = PIC_ADDR;

int mem[100000];

void usage(void)
{
    printf("usage: qlink file [file...]\n");
    exit(1);
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *iofile = fopen(fname, mode);

    if (!iofile)
    {
        printf("Couldn't open %s\n", fname);
        exit(1);
    }

    return iofile;
}

int ReadObject(FILE *infile)
{
    unsigned char len;
    int num, size;

    while (1)
    {
        if (numsym > MAX_SYMBOLS)
        {
            printf("ERROR: Too many symbols\n");
            exit(1);
        }
        if (fread(&type[numsym], 1, 1, infile) != 1) return 0;
        if (type[numsym] == 'E')
        {
            fread(&size, 1, 4, infile);
            break;
        }
        fread(&value[numsym], 1, 4, infile);
        if (addr) value[numsym] += addr - 0x400;
        fread(&len, 1, 1, infile);
        fread(bufptr, 1, len, infile);
        name[numsym] = bufptr;
        bufptr += len;
        if (debugflag)
            printf("%d: %c %8.8x %s\n", numsym, type[numsym], value[numsym], name[numsym]);
        if (!addr && type[numsym] == 'P' && !strcmp(name[numsym], "pictable"))
        {
            if (debugflag)
                printf("Setting pictableaddr = %x\n", value[numsym]);
            pictableaddr = value[numsym];
        }
        numsym++;
    }

    if (addr)
    {
        fread(&mem[addr>>2], 1, 0x400, infile);
        size -= 0x400;
    }
    num = fread(&mem[addr>>2], 1, size, infile);
    if (num != size)
    {
        printf("ERROR: Expected %d bytes, but only read %d bytes\n", size, num);
    }
    addr += num;
    objnum++;
    return 1;
}

int FindGlobal(char *str)
{
    int i;

    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'G' && !strcmp(str, name[i])) return i;
    }

    return -1;
}

int FindVariable(char *str)
{
    int i;

    //printf("FindVariable: Search for %s\n", str);
    for (i = 0; i < numsym; i++)
    {
        //printf("FindVariable: %d %c %s\n", i, type[i], name[i]);
        if (type[i] == 'V' && !strcmp(str, name[i])) return i;
    }

    return -1;
}

int Resolve(int num)
{
    int i, j;
    int resolved = 0;
    int addr_u, addr_g, value_u, value_r;

    for (i = 0; i < num; i++)
    {
        if (type[i] == 'U')
        {
#if 0
            if (debugflag)
                printf("Resolving %s at %8.8x\n", name[i], value[i]);
#endif
            j = FindGlobal(name[i]);
            if (j >= 0)
            {
                if (debugflag)
                    printf("Found %s at %x. Resolving reference at %x\n", name[j], value[j], value[i]);
                addr_u = value[i];
                addr_g = value[j];
                value_u = mem[addr_u>>2];
                value_r = (value_u & ~0xfffff) | ((addr_g - addr_u - 4) & 0xfffff);
                mem[addr_u>>2] = value_r;
                type[i] = 'R';
                resolved++;
            }
#if 0
            else
                printf("Couldn't find global %s\n", name[i]);
#endif
        }
    }
    return resolved;
}

void CountVariables(void)
{
    int i;

    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'V') varcount++;
    }
}

void MergeVariables(int prev_num)
{
    int i, j;

    for (i = prev_num; i < numsym; i++)
    {
        //printf("MergeVariables: Checking %s\n", name[i]);
        if (type[i] == 'V')
        {
            j = FindVariable(name[i]);
#if 0
            if (debugflag)
            {
                printf("Found variable %s at %8.8x\n", name[i], value[i]);
                printf("i = %d, j = %d\n", i, j);
            }
#endif
            if (j < 0 || j == i)
            {
                int oldval = value[i];
                value[i] = (((pictableaddr >> 2) + varcount++) << 20) | (oldval & 0xfffff);
                if (debugflag)
                {
                    printf("New variable %s with value %8.8x.", name[i], oldval);
                    if (oldval == value[i])
                        printf("\n");
                    else
                        printf("Changing value to %8.8x\n", value[i]);
                }
            }
            else
            {
                if (debugflag)
                {
#if 0
                    printf("Merge with entry %d at address %8.8x\n", j, value[j]);
                    printf("Changing %8.8x to %8.8x\n", value[i], value[j]);
#else
                    printf("Existing variable %s with value %8.8x", name[i], value[i]);
                    printf("Changing value to %8.8x\n", value[j]);
#endif
                }
                value[i] = value[j];
                type[i] = 'X';
            }
        }
    }
}

void FixVariableRef(int prev_num)
{
    int i, j;

    for (i = prev_num; i < numsym; i++)
    {
        //printf("FixVariableRef: Checking %s\n", name[i]);
        if (type[i] == 'W')
        {
            j = FindVariable(name[i]);
            if (debugflag)
            {
                printf("Found variable %s at %8.8x. ", name[i], value[i]);
                //printf("i = %d, j = %d\n", i, j);
            }
            if (j < 0)
            {
                if (debugflag)
                    printf("ERROR: Couldn't find variable %s\n", name[i]);
            }
            else
            {
                int addr = value[i];
                int oldval = mem[addr>>2] & 0x1ff;
                int newval = (value[j] >> 20) & 0x1ff;
                if (debugflag)
                    printf("Changing reference at %8.8x from %3.3x to %3.3x\n", addr, oldval, newval);
                oldval = mem[addr>>2];
                newval = (oldval & ~0x1ff) | newval;
#if 0
                if (debugflag)
                    printf("Changing %8.8x to %8.8x\n", oldval, newval);
#endif
                mem[addr>>2] = newval;
            }
        }
    }
}

void AddPicTable(void)
{
    int i;

    if (debugflag) printf("AddPicTable: Setting PIC table address to %8.8x at %8.8x\n", addr, pictableaddr);
    mem[pictableaddr>>2] = addr;
        
    if (debugflag) printf("Number of variables = %d\n", varcount);
    mem[addr>>2] = varcount;
    addr += 4;
    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'V')
        {
            if (debugflag)
                printf("%8.8x %s\n", value[i], name[i]);
            mem[addr>>2] = value[i] & 0xfffff;
            addr += 4;
        }
    }
}

void AddAdjTable(void)
{
    int i;
    int num = 0;
    int adjust_addr = addr;
    int adjtableaddr = pictableaddr - 4;

    if (debugflag) printf("AddAdjTable: Setting ADJ table address to %8.8x at %8.8x\n", addr, adjtableaddr);
    mem[adjtableaddr>>2] = addr;
    addr += 4;
    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'J')
        {
            if (debugflag)
                printf("Adding %8.8x %s to adjustment table\n", value[i] & 0xfffff, name[i]);
            mem[addr>>2] = value[i] & 0xfffff;
            addr += 4;
            num++;
        }
        else if (type[i] == 'V')
        {
            if (!strcmp(name[i], "heapaddr") || !strcmp(name[i], "heapaddrlast"))
            {
                if (debugflag)
                    printf("Adding %8.8x %s to adjustment table\n", value[i] & 0xfffff, name[i]);
                mem[addr>>2] = value[i] & 0xfffff;
                addr += 4;
                num++;
            }
        }
    }
        
    if (debugflag) printf("Number of addresses = %d\n", num);
    mem[adjust_addr>>2] = num;
}

void ReadFile(FILE *infile, int libflag)
{
    int prev_addr = addr;
    int prev_numsym = numsym;
    int prev_varcount = varcount;

    while (ReadObject(infile))
    {
        if (objnum > 1)
        {
            MergeVariables(prev_numsym);
            FixVariableRef(prev_numsym);
        }
        else
            CountVariables();
        if (objnum > 1 && !Resolve(prev_numsym) && libflag)
        {
            addr = prev_addr;
            numsym = prev_numsym;
            varcount = prev_varcount;
            objnum--;
        }
        else
        {
            prev_addr = addr;
            prev_numsym = numsym;
            prev_varcount = varcount;
        }
    }
}

int is_lib(char *str)
{
    int len = strlen(str);
    if (len < 2) return 0;
    if (str[len-1] != 'a') return 0;
    if (str[len-2] != '.') return 0;
    return 1;
}

void InitHeapAddress(void)
{
    int i, hub_addr;
    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'V')
        {
            if (!strcmp(name[i], "heapaddr") || !strcmp(name[i], "heapaddrlast"))
            {
                hub_addr = value[i] & 0xfffff;
                printf("Setting value of %s at location %x to %x\n", name[i], hub_addr, addr);
                mem[hub_addr>>2] = addr;
            }
        }
    }
}

int main(int argc, char **argv)
{
    int i;
    //char buffer[256];
    char outfname[100];
    int unresolved = 0;
    FILE *infile, *outfile;

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            if (!strcmp(argv[i], "-d")) debugflag = 1;
            else usage();
        }
        else
        {
            if (!addr) strcpy(outfname, argv[i]);
            infile = OpenFile(argv[i], "rb");
            ReadFile(infile, is_lib(argv[i]));
            fclose(infile);
        }
    }

    if (!addr) usage();

    Resolve(numsym);
    // Check if any remaining unresolves
    for (i = 0; i < numsym; i++)
    {
        if (type[i] == 'U')
        {
            unresolved++;
            printf("%s is unresolved\n", name[i]);
        }
    }

    if (unresolved)
    {
        printf("%d symbol(s) are unresolved\n", unresolved);
        exit(1);
    }

    AddPicTable();

    AddAdjTable();

    InitHeapAddress();

    for (i = 0; outfname[i]; i++)
    {
        if (outfname[i] == '.') break;
    }
    outfname[i] = 0;
    strcat(outfname, ".bin");
    outfile = OpenFile(outfname, "wb");
    fwrite(mem, 1, addr, outfile);

    return 0;
}
