'*******************************
'*  NTSC 256 x 192 x 8bpp-lut  *
'*******************************

CON

  f_color	= 3_579_545.0		'colorburst frequency
  f_scanline	= f_color / 227.5	'scanline frequency
  f_pixel	= f_scanline * 400.0	'pixel frequency for 400 pixels per scanline

  f_clock	= 50_000_000.0		'clock frequency

  f_xfr		= f_pixel / f_clock * float($7FFF_FFFF)
  f_csc		= f_color / f_clock * float($7FFF_FFFF) * 2.0

  s		= 84			'scales DAC output (s = 0..128)
  r		= s * 78 / 128		'adjusts for modulator expansion

  mody		= ((+38*s/128) & $FF) << 24 + ((+75*s/128) & $FF) << 16 + ((+15*s/128) & $FF) << 8 + (110*s/128 & $FF)
  modi		= ((+76*r/128) & $FF) << 24 + ((-35*r/128) & $FF) << 16 + ((-41*r/128) & $FF) << 8 + (100*s/128 & $FF)
  modq		= ((+27*r/128) & $FF) << 24 + ((-67*r/128) & $FF) << 16 + ((+40*r/128) & $FF) << 8 + 128


DAT		org
'
'
' Setup
'
		rdfast	#0,##$1000-$400		'load .bmp palette into lut
		rep	@.end,#$100
		rflong	y
		shl	y,#8
		wrlut	y,x
		add	x,#1
.end
		rdfast	##256*192/64,##$1000	'set rdfast to wrap on bitmap

		setxfrq ##round(f_xfr)		'set transfer frequency
		setcfrq	##round(f_csc)		'set colorspace converter frequency

		setcy	##mody			'set colorspace converter coefficients
		setci	##modi
		setcq	##modq

		setcmod	#%11_1_0000		'set colorspace converter to YIQ mode (composite)
'
'
' Field loop
'
field           mov	x,#35			'top blanks
		call	#blank

                mov     x,#192			'set visible lines
line	        call	#hsync			'do horizontal sync
		xcont	m_rf,#0			'visible line
		djnz    x,#line           	'another line?

                mov     x,#27			'bottom blanks
		call	#blank

		mov	x,#6			'high vertical syncs
		call	#vsynch

		mov	x,#6			'low vertical syncs
		call	#vsyncl

		mov	x,#6			'high vertical syncs
		call	#vsynch

                jmp     #field                  'loop
'
'
' Subroutines
'
blank		call	#hsync			'blank lines
		xcont	m_vi,#0
		djnz	x,#blank
		ret

hsync		xcont	m_bs,#1			'horizontal sync
		xcont	m_sn,#2
		xcont	m_bc,#1
		xcont	m_cb,c_cb
		xcont	m_bv,#1
		ret

vsynch		xcont	m_bs,#1			'vertical sync high
		xcont	m_hs,#2
		xcont	m_hl,#1
		djnz	x,#vsynch
		ret

vsyncl		xcont	m_bs,#1			'vertical sync low
		xcont	m_hl,#2
		xcont	m_hs,#1
		djnz	x,#vsyncl
		ret
'
'
' Initialized data
'
m_bs		long	$CF000000+50		'before sync
m_sn		long	$CF000000+29		'sync
m_bc		long	$CF000000+7		'before colorburst
m_cb		long	$CF000000+18		'colorburst
m_bv		long	$CF000000+40		'before visible
m_vi		long	$CF000000+256		'visible

m_rf		long	$7F000000+256		'visible rflong 8bpp lut

m_hs		long	$CF000000+20		'vertical sync short
m_hl		long	$CF000000+130		'vertical sync long

c_cb		long	$507000_01		'colorburst reference color
c_vw		long	$FFFF00_00		'white
c_vb		long	$000000_00		'black

x		res	1
y		res	1

'
' Bitmap
'
		orgh	$1000 - $436	'justify pixels at $1000, pallete at $1000-$400
		file	"bitmap.bmp"	'256 x 192, 8bpp-lut
