'*******************************
'*  NTSC 256 x 192 x 8bpp-lut  *
'*      Interrupt-driven       *
'*******************************

CON

  f_color	= 3_579_545.0		'colorburst frequency
  f_scanline	= f_color / 227.5	'scanline frequency
  f_pixel	= f_scanline * 400.0	'pixel frequency for 400 pixels per scanline

  f_clock	= 50_000_000.0		'clock frequency

  f_xfr		= f_pixel / f_clock * float($7FFF_FFFF)
  f_csc		= f_color / f_clock * float($7FFF_FFFF) * 2.0

  s		= 84			'scales DAC output (s = 0..128)
  r		= s * 78 / 128		'adjusts for modulator expansion

  mody		= ((+38*s/128) & $FF) << 24 + ((+75*s/128) & $FF) << 16 + ((+15*s/128) & $FF) << 8 + (110*s/128 & $FF)
  modi		= ((+76*r/128) & $FF) << 24 + ((-35*r/128) & $FF) << 16 + ((-41*r/128) & $FF) << 8 + (100*s/128 & $FF)
  modq		= ((+27*r/128) & $FF) << 24 + ((-67*r/128) & $FF) << 16 + ((+40*r/128) & $FF) << 8 + 128


DAT		org
'
'
' Setup
'
		rdfast	#0,##$1000-$400		'load .bmp palette into lut
		rep	@.end,#$100
		rflong	y
		shl	y,#8
		wrlut	y,x
		add	x,#1
.end
		rdfast	##256*192/64,##$1000	'set rdfast to wrap on bitmap

		setxfrq ##round(f_xfr)		'set transfer frequency
		setcfrq	##round(f_csc)		'set colorspace converter frequency

		setcy	##mody			'set colorspace converter coefficients
		setci	##modi
		setcq	##modq

		setcmod	#%11_1_0000		'set colorspace converter to YIQ mode (composite)

		mov	ijmp1,#field		'set up streamer-empty interrupt
		setint1	#9

		xcont	#10,#0			'do streamer instruction to start interrupt sequence
'
'
' Main program
'
		or	dira,#1			'make p0 output

.loop		xor	outa,#1			'keep toggling p0
		jmp	#.loop
'
'
' Field loop via interrupts - issue next streamer command and then resume
'
field		mov     x,#27+192+35		'set blank+visible+blank lines
.line	        xcont	m_bs,#1			'horizontal sync
		resi1
		xcont	m_sn,#2
		resi1
		xcont	m_bc,#1
		resi1
		xcont	m_cb,c_cb
		resi1
		xcont	m_bv,#1
		resi1
		cmp	x,#27+192+1	wc	'blank line or visible line?
	if_c	cmpr	x,#27		wc
	if_nc	xcont	m_vi,#0			'blank line
	if_c	xcont	m_rf,#0			'visible line
		resi1
		djnz    x,#.line

		mov	x,#6			'high vertical syncs
.vlow		xcont	m_bs,#1
		resi1
		xcont	m_hs,#2
		resi1
		xcont	m_hl,#1
		resi1
		djnz	x,#.vlow

		mov	x,#6			'low vertical syncs
.vhigh		xcont	m_bs,#1
		resi1
		xcont	m_hl,#2
		resi1
		xcont	m_hs,#1
		resi1
		djnz	x,#.vhigh

		mov	x,#6			'high vertical syncs
.vlow2		xcont	m_bs,#1
		resi1
		xcont	m_hs,#2
		resi1
		xcont	m_hl,#1
		resi1
		djnz	x,#.vlow2

                jmp     #field                  'loop
'
'
' Initialized data
'
m_bs		long	$CF000000+50		'before sync
m_sn		long	$CF000000+29		'sync
m_bc		long	$CF000000+7		'before colorburst
m_cb		long	$CF000000+18		'colorburst
m_bv		long	$CF000000+40		'before visible
m_vi		long	$CF000000+256		'visible

m_rf		long	$7F000000+256		'visible rflong 8bpp lut

m_hs		long	$CF000000+20		'vertical sync short
m_hl		long	$CF000000+130		'vertical sync long

c_cb		long	$507000_01		'colorburst reference color
c_vw		long	$FFFF00_00		'white
c_vb		long	$000000_00		'black

x		res	1
y		res	1

'
' Bitmap
'
		orgh	$1000 - $436	'justify pixels at $1000, pallete at $1000-$400
		file	"bitmap.bmp"
