con
        sys_clk = 50_000_000
        baud_rate = 115_200
        
        rx_pin = 63
        tx_pin = 62
        
        rx_reg = $FFF80
        tx_reg = $FFF84

        edg_int = 5
        wrl_int = 7
        
dat
                orgh
                org 0

                setwrl  #(rx_reg - $FFF80) >> 2         ' set wrl-event the receive buffer
                
                call    #start_fds

wait4rx         waitwrl                                 ' wait for a long to be written to receive buffer

                rdlong  .buff, ##rx_reg                 ' read from RX buffer
                wrlong  .buff, ##tx_reg                 ' echo to TX buffer
                
                jmp     #wait4rx
                
.buff           long    0               
                
start_fds       coginit #16, #@fds
                ret

'==================================================================================================
                org 0
fds             
                setb    outb, #tx_pin                   ' set TX high (inactive)
                setb    dirb, #tx_pin                   ' configure TX pin
                clrb    dirb, #rx_pin                   ' configure RX pin

                mov     adra, #transmit                 ' entrypoint to transmit co-routine
                mov     adrb, #receive                  ' entrypoint to receive co-routine

                setedg  #(%0_10_000000 | rx_pin)        ' set up for falling edge
                mov     ijmp1, #rx_isr                  ' set ISR
                setint1 #edg_int                        ' enable interrupt
                
                setwrl  #(tx_reg-$FFF80) >> 2           ' set up for WRLONG
                
                mov     ijmp2, #tx_isr                  ' set ISR
                setint2 #wrl_int                        ' enable interrupt
                
main_loop       waitint                                 ' wait for one of the interrupts to occur
                
.loop_rxtx
.check_tx       tjz     tx_buff, #.check_rx             ' if not transmitting, skip over tx call
                pollct2 wc
        if_c    calld   adra, adra wc, wz               ' call TX co-routine

.check_rx       tjz     rx_buff, #.post_rxtx            ' if not receiving, skip over rx call
                pollct1 wc
        if_c    calld   adrb, adrb wc, wz               ' call RX co-routine
        
.post_rxtx      tjnz    rx_buff, #.loop_rxtx
                tjnz    tx_buff, #.loop_rxtx
                jmp     #main_loop

'--------------------------------------------------------------------------------------------------
transmit_rep    calld   adra, adra wc, wz               ' switch back to main task
transmit        shr     tx_buff, #1 wc                  ' move next bit to C
                setbc   outb, #tx_pin                   ' set pin to C
                tjz     tx_buff, #.tx_end               ' check to see if buffer is empty
                addct2  tx_cnt, full_bit_time           ' not empty, so prepare for next bit time
                jmp     #transmit_rep                   ' jump to CALLD

.tx_end         push    adra                            ' push return address to stack
                mov     adra, #transmit                 ' reset co-routine address
                ret                                     ' return (pop from stack)
                
'--------------------------------------------------------------------------------------------------
receive         jnp     #rx_pin, #.good_start
                mov     rx_buff, ##-1                   ' bad start bit
                jmp     #.rx_end

.good_start     addct1  rx_cnt, full_bit_time           ' get the next sample time

.read_bit       calld   adrb, adrb wc, wz               ' switch back to main task
                testb   inb, #rx_pin wc                 ' get pin state to C
                testb   rx_buff, #8 wz                  ' get buffer state to Z
        
        if_nz   jmp     #.check_end                     ' buffer is full (z=1)

.store_bit      rcl     rx_buff, #1                     ' store the bit
                addct1  rx_cnt, full_bit_time           ' get next sample time
                jmp     #.read_bit                      ' jump to CALLD

.check_end
        if_nc   mov     rx_buff, ##-2                   ' missing stop bit (c=0)
        if_c    rev     rx_buff                         ' since we received LSB first, reverse the bits
        if_c    shr     rx_buff, #24                    ' and shift back to 7..0
                jmp     #.rx_end

.rx_end         wrlong  rx_buff, ##rx_reg               ' store value to hub
                mov     rx_buff, #0                     ' clear (which indicates read is no longer active)

                push    adrb                            ' push return address to stack
                mov     adrb, #receive                  ' reset co-routine address
                ret                                     ' return (pop from stack)

'--------------------------------------------------------------------------------------------------
rx_isr          tjnz    rx_buff, #.ret                  ' see if we are already receiving a byte
                                                        ' buffer is non-zero. this is not the start bit.
                getct   rx_cnt
                addct1  rx_cnt, half_bit_time           ' add 0.5 bit periods (to test valid start bit)
                
                mov     rx_buff, #1                     ' set rx_buff to non-zero.  This bit will
                                                        ' be shifted out of the lower byte. It will be
                                                        ' used to detect when a full byte is received.

.ret            reti1

'--------------------------------------------------------------------------------------------------
tx_isr          tjnz    tx_buff, #.ret                  ' see if we are currently sending a byte
                                                        ' buffer is non-zero. do not overwrite.

                rdlong  tx_buff, ##tx_reg               ' get the byte to be transmitted
                
                and     tx_buff, #$FF                   ' keep only the 8 LSBs
                setb    tx_buff, #8                     ' set tx_buff[8] to stop bit (also ensures buffer
                                                        ' is non-zero)

                clrb    outb, #tx_pin                   ' set pin low (start bit)

                getct   tx_cnt
                addct2  tx_cnt, full_bit_time           ' add 1.0 bit periods for start bit

.ret            reti2

'--------------------------------------------------------------------------------------------------
full_bit_time   long    sys_clk / baud_rate
half_bit_time   long    (sys_clk / baud_rate) >> 1

rx_buff         res     1
tx_buff         res     1

rx_cnt          res     1
tx_cnt          res     1

t1              res     1