con
        sys_clk = 50_000_000
        baud_rate = 115_200
        
        rx_pin = 63
        tx_pin = 62
        
        rx_reg = $FFF80
        tx_reg = $FFF84

        ct1_int = 1
        ct2_int = 2
        edg_int = 5
        wrl_int = 7
        
dat
                orgh
                org 0

                setwrl  #(rx_reg - $FFF80) >> 2         ' set wrl-event the receive buffer
                
                call    #start_fds

                waitx   ##50_000
                wrlong #"o", ##tx_reg
                waitx   ##50_000
                wrlong #"w", ##tx_reg
                waitx   ##50_000
                wrlong #"N", ##tx_reg
                waitx   ##50_000
                wrlong #"w", ##tx_reg

wait4rx         waitwrl                                 ' wait for a long to be written to receive buffer

                rdlong  .buff, ##rx_reg                 ' read from RX buffer
                wrlong  .buff, ##tx_reg                 ' echo to TX buffer
                
                jmp     #wait4rx
                
.buff           long    0               
                
start_fds       coginit #16, #@fds
                ret

'==================================================================================================
                org 0
fds             
                setb    outb, #tx_pin                   ' set TX high (inactive)
                setb    dirb, #tx_pin                   ' configure TX pin
                clrb    dirb, #rx_pin                   ' configure RX pin

                mov     ijmp1, #rx_isr                  ' ISR for RX start
                mov     ijmp2, #tx_isr                  ' ISR for TX timer
                mov     ijmp3, #wr_isr                  ' ISR for TX start
                
                setedg  #(%0_10_000000 | rx_pin)        ' set up for falling edge
                setint1 #edg_int                        ' enable interrupt
                
                setwrl  #(tx_reg-$FFF80) >> 2           ' set up for WRLONG
                setint3 #wrl_int                        ' enable interrupt
                
rx_loop         waitint                                 ' wait for one of the interrupts to occur
                nop                                     ' since TJZ depends on ISR, NOP to make sure
                                                        ' TJZ executes after return from ISR. If NOP
                                                        ' is removed, TJZ would execute before ISR.
                tjz     rx_buff, #rx_loop               ' if not receiving, go back to waiting

.check_start    pollct1 wc
        if_nc   jmp     #.check_start

                jnp     #rx_pin, #.good_start           ' pin is low, good start
                mov     rx_buff, ##-1                   ' bad start bit
                jmp     #.rx_end

.good_start     addct1  rx_cnt, full_bit_time           ' get the first bit sample time

.read_bit       pollct1 wc
        if_nc   jmp     #.read_bit

                testb   inb, #rx_pin wc                 ' get pin state to C

                testb   rx_buff, #8 wz                  ' get buffer state to Z 
        if_nz   jmp     #.check_end                     ' buffer is full (z=1)

                rcl     rx_buff, #1                     ' store the bit
                addct1  rx_cnt, full_bit_time           ' get the next bit sample time
                jmp     #.read_bit

.check_end
        if_nc   mov     rx_buff, ##-2                   ' missing stop bit (c=0)
        if_c    rev     rx_buff                         ' since we received LSB first, reverse the bits
        if_c    getbyte rx_buff, rx_buff, #3            ' and get the byte (which is at 31..24 due to the REV)

.rx_end         wrlong  rx_buff, ##rx_reg               ' store value to hub
                mov     rx_buff, #0                     ' clear (which indicates read is no longer active)

                jmp     #rx_loop

'--------------------------------------------------------------------------------------------------
rx_isr          tjnz    rx_buff, #.ret                  ' see if we are already receiving a byte
                                                        ' buffer is non-zero. this is not the start bit.
                getct   rx_cnt
                addct1  rx_cnt, half_bit_time           ' add 0.5 bit periods (to test valid start bit)
                
                mov     rx_buff, #1                     ' set rx_buff to non-zero.  This bit will
                                                        ' be shifted out of the lower byte. It will be
                                                        ' used to detect when a full byte is received.
.ret            reti1

'--------------------------------------------------------------------------------------------------
wr_isr          tjnz    tx_buff, #.ret
                rdlong  tx_buff, ##tx_reg               ' get the byte to be transmitted
                
                and     tx_buff, #$FF                   ' keep only the 8 LSBs
                setb    tx_buff, #8                     ' set tx_buff[8] to stop bit (also ensures buffer
                                                        ' is non-zero)
                getct   tx_cnt
                clrb    outb, #tx_pin                   ' set pin low (start bit)

                addct2  tx_cnt, full_bit_time           ' add 1.0 bit periods for start bit
                setint2 #ct2_int

.ret            reti3

'--------------------------------------------------------------------------------------------------
tx_isr          shr     tx_buff, #1 wc                  ' move next bit to C
                setbc   outb, #tx_pin                   ' set pin to C
                tjz     tx_buff, #.tx_end               ' check to see if buffer is empty
                addct2  tx_cnt, full_bit_time           ' not empty, so prepare for next bit time
                reti2

.tx_end         setint2 #0                              ' disable interrupt
                reti2

'--------------------------------------------------------------------------------------------------
full_bit_time   long    sys_clk / baud_rate
half_bit_time   long    (sys_clk / baud_rate) >> 1

rx_buff         res     1
tx_buff         res     1

rx_cnt          res     1
tx_cnt          res     1

t1              res     1
